package com.a.eye.skywalking.trace;

import com.a.eye.skywalking.messages.ISerializable;
import com.a.eye.skywalking.trace.TraceId.DistributedTraceId;
import com.a.eye.skywalking.trace.TraceId.NewDistributedTraceId;
import com.a.eye.skywalking.trace.TraceId.PropagatedTraceId;
import com.a.eye.skywalking.trace.proto.SegmentMessage;
import com.a.eye.skywalking.trace.proto.SegmentRefMessage;
import com.a.eye.skywalking.trace.proto.SpanMessage;
import com.google.protobuf.ProtocolStringList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * {@link TraceSegment} is a segment or fragment of the distributed trace.
 * {@see https://github.com/opentracing/specification/blob/master/specification.md#the-opentracing-data-model}
 * A {@link
 * TraceSegment} means the segment, which exists in current {@link Thread}. And the distributed trace is formed by multi
 * {@link TraceSegment}s, because the distributed trace crosses multi-processes, multi-threads.
 *
 * Created by wusheng on 2017/2/17.
 */
public class TraceSegment implements ISerializable<SegmentMessage> {
    private static final String ID_TYPE = "Segment";

    /**
     * The id of this trace segment.
     * Every segment has its unique-global-id.
     */
    private String traceSegmentId;

    /**
     * The start time of this trace segment.
     */
    private long startTime;

    /**
     * The end time of this trace segment.
     */
    private long endTime;

    /**
     * The refs of parent trace segments, except the primary one.
     * For most RPC call, {@link #refs} contains only one element,
     * but if this segment is a start span of batch process, the segment faces multi parents,
     * at this moment, we use this {@link #refs} to link them.
     */
    private List<TraceSegmentRef> refs;

    /**
     * The spans belong to this trace segment.
     * They all have finished.
     * All active spans are hold and controlled by "skywalking-api" module.
     */
    private List<Span> spans;

    /**
     * The <code>applicationCode</code> represents a name of current application/JVM and indicates which is business
     * role in the cluster.
     *
     * e.g. account_app, billing_app
     */
    private String applicationCode;

    /**
     * The <code>relatedGlobalTraces</code> represent a set of all related trace. Most time it contains only one
     * element, because only one parent {@link TraceSegment} exists, but, in batch scenario, the num becomes greater
     * than 1, also meaning multi-parents {@link TraceSegment}.
     *
     * The difference between <code>relatedGlobalTraces</code> and {@link #refs} is:
     * {@link #refs} targets this {@link TraceSegment}'s direct parent,
     *
     * and
     *
     * <code>relatedGlobalTraces</code> targets this {@link TraceSegment}'s related call chain, a call chain contains
     * multi {@link TraceSegment}s, only using {@link #refs} is not enough for analysis and ui.
     */
    private LinkedList<DistributedTraceId> relatedGlobalTraces;

    /**
     * Create a trace segment, by given segmentId.
     * This segmentId is generated by TraceSegmentRef, AKA, from tracer/agent module.
     */
    public TraceSegment(String applicationCode) {
        this.traceSegmentId = GlobalIdGenerator.generate(ID_TYPE);
        this.applicationCode = applicationCode;
        this.startTime = System.currentTimeMillis();
        this.spans = new LinkedList<Span>();
        this.relatedGlobalTraces = new LinkedList<DistributedTraceId>();
        this.relatedGlobalTraces.add(new NewDistributedTraceId());
    }

    /**
     * Create a trace segment, by given {@link SegmentMessage}
     *
     * @param message from another {@link TraceSegment#serialize()}
     */
    public TraceSegment(SegmentMessage message) {
        deserialize(message);
    }

    /**
     * Establish the link between this segment and its parents.
     *
     * @param refSegment {@link TraceSegmentRef}
     */
    public void ref(TraceSegmentRef refSegment) {
        if(!refs.contains(refSegment)){
            refs.add(refSegment);
        }
    }

    public void relatedGlobalTraces(List<DistributedTraceId> distributedTraceIds) {
        if (distributedTraceIds == null || distributedTraceIds.size() == 0) {
            return;
        }
        if (relatedGlobalTraces.getFirst() instanceof NewDistributedTraceId) {
            relatedGlobalTraces.removeFirst();
        }
        for (DistributedTraceId distributedTraceId : distributedTraceIds) {
            if(!relatedGlobalTraces.contains(distributedTraceId)){
                relatedGlobalTraces.add(distributedTraceId);
            }
        }
    }

    /**
     * After {@link Span} is finished, as be controller by "skywalking-api" module,
     * notify the {@link TraceSegment} to archive it.
     *
     * @param finishedSpan
     */
    public void archive(Span finishedSpan) {
        spans.add(finishedSpan);
    }

    /**
     * Finish this {@link TraceSegment}.
     *
     * return this, for chaining
     */
    public TraceSegment finish() {
        this.endTime = System.currentTimeMillis();
        return this;
    }

    public String getTraceSegmentId() {
        return traceSegmentId;
    }

    public long getStartTime() {
        return startTime;
    }

    public long getEndTime() {
        return endTime;
    }

    public List<TraceSegmentRef> getRefs() {
        if (refs == null) {
            return null;
        }
        return Collections.unmodifiableList(refs);
    }

    public List<DistributedTraceId> getRelatedGlobalTraces() {
        return Collections.unmodifiableList(relatedGlobalTraces);
    }

    public List<Span> getSpans() {
        return Collections.unmodifiableList(spans);
    }

    public String getApplicationCode() {
        return applicationCode;
    }

    @Override public String toString() {
        return "TraceSegment{" +
            "traceSegmentId='" + traceSegmentId + '\'' +
            ", startTime=" + startTime +
            ", endTime=" + endTime +
            ", refs=" + refs +
            ", spans=" + spans +
            ", applicationCode='" + applicationCode + '\'' +
            ", relatedGlobalTraces=" + relatedGlobalTraces +
            '}';
    }

    @Override
    public SegmentMessage serialize() {
        SegmentMessage.Builder segmentBuilder = SegmentMessage.newBuilder();
        segmentBuilder.setTraceSegmentId(traceSegmentId);
        segmentBuilder.setStartTime(startTime);
        segmentBuilder.setEndTime(endTime);
        segmentBuilder.setApplicationCode(applicationCode);
        if (refs != null && refs.size() > 0) {
            for (TraceSegmentRef ref : refs) {
                segmentBuilder.addRefs(ref.serialize());
            }
        }
        for (DistributedTraceId id : relatedGlobalTraces) {
            segmentBuilder.addRelatedTraceIds(id.get());
        }
        for (Span span : spans) {
            segmentBuilder.addSpans(span.serialize());
        }
        return segmentBuilder.build();
    }

    @Override
    public void deserialize(SegmentMessage message) {
        traceSegmentId = message.getTraceSegmentId();
        startTime = message.getStartTime();
        endTime = message.getEndTime();
        applicationCode = message.getApplicationCode();
        List<SegmentRefMessage> refsList = message.getRefsList();
        if (refsList != null && refsList.size() > 0) {
            this.refs = new LinkedList<TraceSegmentRef>();
            for (SegmentRefMessage refMessage : refsList) {
                TraceSegmentRef ref = new TraceSegmentRef();
                ref.deserialize(refMessage);
                refs.add(ref);
            }
        }
        ProtocolStringList relatedTraceIdsList = message.getRelatedTraceIdsList();
        this.relatedGlobalTraces = new LinkedList<DistributedTraceId>();
        for (String id : relatedTraceIdsList) {
            relatedGlobalTraces.add(new PropagatedTraceId(id));
        }

        List<SpanMessage> spansList = message.getSpansList();
        if (spansList != null) {
            this.spans = new LinkedList<Span>();
            for (SpanMessage spanMessage : spansList) {
                spans.add(new Span(spanMessage));
            }
        }
    }
}
